var babelGenerator = require("@babel/generator");
var babelParser = require("@babel/parser");
var SlJsInfra = require("sl-js-infra").SlJsInfra;
var astTypes = require("ast-types");

var plugins = ["*", ["decorators", {decoratorsBeforeExport: true}]];
var updateStatementCoverageCommand = ".s\\[\\d+\\]\\++";
var updateBranchCoverageCommand = ".b\\[\\d+\\]\\[\\d+\\]\\++";
var propNameToShort = {
    start: "st",
    end: "e",
    line: "l",
    column: "c",
    name: "n",
    decl: "d",
    loc: "lc",
    type: "t",
    locations: "lcs"
};


function reduceContentSize(filename, instrumentedCode, babylonPlugins, shouldRemoveBranchCoverage, shouldMinifyOutput, logger) {
    try {
        var ast = createAST(instrumentedCode, babylonPlugins);
        var coverageObjectNode = getCoverageObjectNode(ast);
        var coverageObjectName = coverageObjectNode.id.name;
        reduceCoverageObjectSize(coverageObjectNode, shouldRemoveBranchCoverage);
        var updateStatementCoverageRegEx = new RegExp(coverageObjectName + updateStatementCoverageCommand);
        var updateBranchCoverageRegEx = new RegExp(coverageObjectName + updateBranchCoverageCommand);
        removeRedundantUpdateCommands(ast, shouldRemoveBranchCoverage, updateStatementCoverageRegEx, updateBranchCoverageRegEx);
        var finalContent = parseASTToCode(ast, shouldMinifyOutput);
        validateFinalContent(finalContent, shouldRemoveBranchCoverage, updateStatementCoverageRegEx, updateBranchCoverageRegEx);
        return finalContent
    } catch (e) {
        var message = "Error while removing statement coverage from '" + filename + "' Error: '" + e + "'";
        SlJsInfra.CockpitNotifier.sendError(message);
        logger.warn(message + " Using file with statement coverage data");
        return instrumentedCode;
    }
}

function verifyUpdateCoverageRemoved(finalContent, updateStatementCoverageRegEx, coverageElement) {
    var match = finalContent.match(updateStatementCoverageRegEx);
    if (match != null) {
        throw new Error("Expected all " + coverageElement + " related expression to be removed")
    }
}

function validateFinalContent(finalContent, shouldRemoveBranchCoverage, updateStatementCoverageRegEx, updateBranchCoverageRegEx) {
    verifyUpdateCoverageRemoved(finalContent, updateStatementCoverageRegEx, "statement");
    if(shouldRemoveBranchCoverage) {
        verifyUpdateCoverageRemoved(finalContent, updateBranchCoverageRegEx, "branch");
    }
}


function shouldConvertToEmptyObject(path, shouldRemoveBranchCoverage) {
    return isStatementCoverageRelatedObject(path.value) || (shouldRemoveBranchCoverage && isBranchCoverageRelatedObject(path.value));
}

function shouldUseShortPropNames(path, shouldRemoveBranchCoverage) {
    if (path && path.key && path.key.name){
        return path.key.name == "fnMap" || (!shouldRemoveBranchCoverage && path.key.name == "branchMap")
    }
    return isStatementCoverageRelatedObject(path.value) || (shouldRemoveBranchCoverage && isBranchCoverageRelatedObject(path.value));
}

function nodeShouldBeDeleted(path, updateStatementCoverageRegEx, shouldRemoveBranchCoverage, updateBranchCoverageRegEx) {
    return isStatementCoverageUpdateExpresion(path.node, updateStatementCoverageRegEx) || (shouldRemoveBranchCoverage && isBranchCoverageUpdateExpresion(path.node, updateBranchCoverageRegEx));
}

/**
 * Traverse over the ast and deletes statement (and branch) coverage update  commands (cov_1234.s[1]++)
 * @param ast
 * @param shouldRemoveBranchCoverage
 * @param updateStatementCoverageRegEx
 * @param updateBranchCoverageRegEx
 */
function removeRedundantUpdateCommands(ast, shouldRemoveBranchCoverage, updateStatementCoverageRegEx, updateBranchCoverageRegEx) {
    astTypes.visit(ast, {
        visitUpdateExpression(path) {
            if (nodeShouldBeDeleted(path, updateStatementCoverageRegEx, shouldRemoveBranchCoverage, updateBranchCoverageRegEx)) {
                path.prune()
            }
            this.traverse(path)
        }
    })
}

/**
 * Traverse over the ast of the coverage object and do the followings:
 * 1. Set 'statementMap' and 's' to empty objects
 * 2. If user not use branchCoverage, remove also set also 'branchMap' and 'b' to empty projects
 * 3. Convert elements metadata (start, end, name, etc...) to short names based on 'propNameToShort'
 * @param ast
 * @param shouldRemoveBranchCoverage
 */
function reduceCoverageObjectSize(ast, shouldRemoveBranchCoverage) {
    astTypes.visit(ast, {
        visitObjectProperty(path) {
            if (shouldConvertToEmptyObject(path, shouldRemoveBranchCoverage)) {
                path.value.value.properties = [];
            } else if (shouldUseShortPropNames(path.value, shouldRemoveBranchCoverage)){
                convertToShortPropNames(path.value)
            }
            this.traverse(path)
        }
    })
}

function convertToShortPropNames(path) {
    astTypes.visit(path, {
        visitObjectProperty(path) {
            if(path && path.value && path.value.key && path.value.key.name){
                path.value.key.name = propNameToShort[path.value.key.name] || path.value.key.name;
            }
            this.traverse(path)
        }
    })
}

function getCoverageObjectNode(ast) {
    var coverageObject;
    astTypes.visit(ast, {
        visitVariableDeclarator(path) {
            if (path.value && path.value.id && path.value.id.name && path.value.id.name.startsWith("cov_")) {
                coverageObject = path.value;
                this.abort();
            }
            this.traverse(path)
        }
    })
    if (!coverageObject) {
        throw new Error("Could find coverage object")
    }
    return coverageObject;
}

function parseASTToCode(ast, minified) {
    try {
        return babelGenerator.default(ast, {
            comments: false,
            compact: true,
            minified: minified
        }, '').code;
    } catch (e) {
        throw e
    }
}

function createAST(instrumentedCode, babylonPlugins) {
    return babelParser.parse(instrumentedCode, {
        sourceType: "unambiguous",
        plugins: plugins.concat(babylonPlugins)
    })
}

function isStatementCoverageRelatedObject(node) {
    return node && node.key && (node.key.name === "statementMap" || node.key.name === "s") && node.value;
}

function isBranchCoverageRelatedObject(node) {
    return node && node.key && (node.key.name === "branchMap" || node.key.name === "b") && node.value;
}

function isStatementCoverageUpdateExpresion(node, updateStatementCoverageRegEx) {
    var expressionCode = parseASTToCode(node, false);
    return updateStatementCoverageRegEx.test(expressionCode)
}

function isBranchCoverageUpdateExpresion(node, updateBranchCoverageRegEx) {
    var expressionCode = parseASTToCode(node, false);
    return updateBranchCoverageRegEx.test(expressionCode)
}

module.exports.reduceContentSize = reduceContentSize;
